@@ -24,7 +24,7 @@
 PlugExport int plugin_init(struct arglist * desc)
 {
  plug_set_id(desc, 10335);
- plug_set_version(desc, "$Revision: 1.12 $");
+ plug_set_version(desc, "$Revision: 1.25 $");
    
          
  plug_set_name(desc, "Nessus TCP scanner", NULL);
@@ -52,11 +52,14 @@ SYN (half open) scanners\
 
 typedef struct {
   int			fd;
-  time_t		tictac;		/* open time */
+  struct timeval	tictac;		/* open time */
   unsigned short	port;
   unsigned char		state;
 } grab_socket_t;
 
+#define DIFFTV(t1,t2)	(t1.tv_sec - t2.tv_sec + (t1.tv_usec - t2.tv_usec) / 1000000)
+#define DIFFTVu(t1,t2)	((t1.tv_sec - t2.tv_sec) * 1000000.0 + (t1.tv_usec - t2.tv_usec))
+
 #define GRAB_SOCKET_UNUSED	0
 #define GRAB_SOCKET_OPENING	1
 #define GRAB_SOCKET_OPEN	2
@@ -71,6 +74,18 @@ typedef struct {
 #ifdef DEBUG
 # define DISPLAY
 #endif
+#undef STATISTICS
+/*
+ * If you want to enable the "statistics", define STATISTICS and link 
+ * the plugin with libm (we need sqrt)
+ */
+#undef BLACKLIST_FILTERED_PORTS
+/*
+ * MA 2005-05-08: Too dangerous, as nessus_tcp_scanner may miss connections
+ * against slow machines.
+ * By the way, nessus_tcp_scanner was the only port scanner that did this.
+ */
+
 
 static int
 my_socket_close(int s)
@@ -85,6 +100,7 @@ my_socket_close(int s)
   return close(s);
 }
 
+#ifdef BLACKLIST_FILTERED_PORTS
 static void
 remember_filtered_port(struct arglist* desc, int port)
 {
@@ -92,6 +108,7 @@ remember_filtered_port(struct arglist* desc, int port)
   snprintf(k, sizeof(k), "/tmp/ConnectTimeout/TCP/%d", port);
   plug_set_key(desc, k, ARG_INT, (void*)1); 
 }
+#endif
 
 static int
 banner_grab(const struct in_addr *pia, const char* portrange, 
@@ -102,22 +119,27 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	    struct arglist *desc,
 	    struct arglist *hostinfos)
 {
-  char			buf[8192];
+  char			buf[2048], kb[64];
   int			s, tcpproto;
   struct protoent	*proto;
   fd_set		rfs, wfs, efs;
-  struct timeval	timeout;
+  struct timeval	timeout, ti;
   struct sockaddr_in	sa;
   int			port = 23;
-  int			imax, i, j, scanned_ports, x, opt, optsz;
+  int			imax, i, j, scanned_ports, x, opt;
+  unsigned int		optsz;
   int			minport;
   unsigned char		ports_states[65536];
   grab_socket_t		sockets[GRAB_MAX_SOCK];
   int			open_sock_nb, open_sock_max, open_sock_max2;
-  time_t		ti;
   int			unfiltered_ports_nb, filtered_ports_nb, timeout_nb;
+  int			open_ports_nb, closed_ports_nb;
   int			untested_ports_nb, total_ports_nb;
-
+#ifdef STATISTICS
+  double		rtt_sum[3], rtt_sum2[3];
+  int			rtt_nb[3], rtt_max[3], rtt_min[3];
+  static const char	*rtt_type[] = {"unfiltered", "open", "closed" };
+#endif
 
   proto = getprotobyname("tcp");
   if (proto == NULL)
@@ -130,6 +152,14 @@ banner_grab(const struct in_addr *pia, const char* portrange,
   for (i = 0; i < sizeof(ports_states) / sizeof(*ports_states); i ++)
     ports_states[i] = GRAB_PORT_NOT_TESTED;
   scanned_ports = 0;
+#ifdef STATISTICS
+  for (i = 0; i < 3; i ++)
+    {
+      rtt_sum[i] = rtt_sum2[i] = 0.0;
+      rtt_max[i] = rtt_nb[i] = 0;
+      rtt_min[i] = 1 << (sizeof(int) * 4) - 1;
+    }
+#endif
 
   {
     char	*p, *q;
@@ -224,7 +254,7 @@ banner_grab(const struct in_addr *pia, const char* portrange,
   open_sock_nb = 0; 
   open_sock_max = min_cnx; open_sock_max2 = max_cnx;
 
-  filtered_ports_nb = unfiltered_ports_nb = 0;
+  open_ports_nb = closed_ports_nb = filtered_ports_nb = unfiltered_ports_nb = 0;
 
   minport = 1;
   while (scanned_ports < 65535)
@@ -316,7 +346,7 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 		  sockets[open_sock_nb].fd = s;
 		  sockets[open_sock_nb].port = port;
 		  sockets[open_sock_nb].state = GRAB_SOCKET_OPENING;
-		  sockets[open_sock_nb].tictac = time(NULL);
+		  (void) gettimeofday(&sockets[open_sock_nb].tictac, NULL);
 		  open_sock_nb ++;
 		  break;
 		  
@@ -332,6 +362,7 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 		  ports_states[port] = GRAB_PORT_CLOSED;
 		  my_socket_close(s);
 		  unfiltered_ports_nb ++;
+		  closed_ports_nb ++;
 		  untested_ports_nb --;
 		  continue;
 		  
@@ -352,10 +383,11 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	      sockets[open_sock_nb].fd = s;
 	      sockets[open_sock_nb].port = port;
 	      sockets[open_sock_nb].state = GRAB_SOCKET_OPEN;
-	      sockets[open_sock_nb].tictac = time(NULL);
+	      (void) gettimeofday(&sockets[open_sock_nb].tictac, NULL);
 	      open_sock_nb ++;
 	      ports_states[port] = GRAB_PORT_OPEN;
 	      unfiltered_ports_nb ++;
+	      open_ports_nb ++;
 	      untested_ports_nb --;
 	      scanner_add_port(desc, port, "tcp");
 	    }
@@ -374,7 +406,7 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	      return -1;
 	    }
 	  else
-	    return 0;
+	    goto end;
 	}
 
       FD_ZERO(&rfs); FD_ZERO(&wfs); FD_ZERO(&efs);
@@ -417,12 +449,16 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 #ifdef DEBUG
 	      fprintf(stderr, "No socket! %d ports remaining\n", untested_ports_nb);
 #endif
-	      return 0;
+	      goto end;
 	    }
 	}
 
       timeout.tv_sec = read_timeout; /* * 2 ? */
-      timeout.tv_usec = 0;
+      /*
+       * Some randomness so that not all scanners fire at the same time
+       * when several firewalled machines are scanned in parallel.
+       */
+      timeout.tv_usec = (unsigned)(lrand48() & 0x7FFFFFFF) % 1000000;
       i = 0;
       do
 	x = select(imax + 1, &rfs, &wfs, NULL, &timeout);
@@ -449,7 +485,9 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 		      ports_states[sockets[i].port] = GRAB_PORT_FILTERED;
 		      filtered_ports_nb ++;
 		      untested_ports_nb --;
+#ifdef BLACKLIST_FILTERED_PORTS
 		      remember_filtered_port(desc, sockets[i].port);
+#endif
 		    }
 		  
 		}
@@ -458,6 +496,7 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	}
       else			/* something to do */
 	{
+	  (void) gettimeofday(&ti, NULL);
 	  for (i = 0; i < open_sock_nb; i ++)
 	    {
 	      if (sockets[i].fd > 0)
@@ -470,17 +509,30 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 			return -1;
 		      }
 
+		    x = DIFFTVu(ti, sockets[i].tictac);
+#if defined DEBUG || defined STATISTICS
+		    fprintf(stderr, "RTT to %s:%d: %g s\n", 
+			    inet_ntoa(*pia), sockets[i].port, x / 1e6);
+#endif
 		    if (opt != 0)
 		      {
 			errno = opt;
 #ifdef DEBUG
 			perror("select->getsockopt");
+#endif
+#ifdef STATISTICS
+			rtt_nb[2] ++;
+			rtt_sum[2] += (double)x;
+			rtt_sum2[2] += (double)x * (double)x;
+			if (x > rtt_max[2]) rtt_max[2] = x;
+			if (x < rtt_min[2]) rtt_min[2] = x;
 #endif
 			ports_states[sockets[i].port] = GRAB_PORT_CLOSED;
 			my_socket_close(sockets[i].fd);
 			sockets[i].fd = -1;
 			sockets[i].state = GRAB_SOCKET_UNUSED;
 			unfiltered_ports_nb ++;
+			closed_ports_nb ++;
 			untested_ports_nb --;
 #ifdef DISPLAY
 			printf(">> %d: CLOSED\n", sockets[i].port);
@@ -492,24 +544,63 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 #ifdef DISPLAY
 			printf(">> %d: OPEN\n", sockets[i].port);
 #endif
+#ifdef STATISTICS
+			rtt_nb[1] ++;
+			rtt_sum[1] += (double)x;
+			rtt_sum2[1] += (double)x * (double)x;
+			if (x > rtt_max[1]) rtt_max[1] = x;
+			if (x < rtt_min[1]) rtt_min[1] = x;
+#endif
+			unfiltered_ports_nb ++;
+			open_ports_nb ++;
 			untested_ports_nb --;
 			ports_states[sockets[i].port] = GRAB_PORT_OPEN;
 			scanner_add_port(desc, sockets[i].port, "tcp");
+			snprintf(kb, sizeof(kb), "TCPScanner/CnxTime1000/%d", sockets[i].port);
+			plug_set_key(desc, kb, ARG_INT, (void*)(x/1000));
+			snprintf(kb, sizeof(kb), "TCPScanner/CnxTime/%d", sockets[i].port);
+			plug_set_key(desc, kb, ARG_INT, (void*)((x + 500000) / 1000000));
+			sockets[i].tictac = ti;
 		      }
+#ifdef STATISTICS
+		    rtt_nb[0] ++;
+		    rtt_sum[0] += (double)x;
+		    rtt_sum2[0] += (double)x * (double)x;
+		    if (x > rtt_max[0]) rtt_max[0] = x;
+		    if (x < rtt_min[0]) rtt_min[0] = x;
+#endif
 		  }
 		else if (FD_ISSET(sockets[i].fd, &rfs))
 		  {
 		    x = read(sockets[i].fd, buf, sizeof(buf)-1);
 		    if (x > 0)
 		      {
-			char	kb[64];
+			char	buf2[sizeof(buf)*2+1];
+			int y, flag = 0;
+
+			for (y = 0; y < x; y ++)
+			  {
+			    sprintf(buf2 + 2*y, "%02x", (unsigned char) buf[y]);
+			    if (buf[y] == '\0') flag = 1;
+			  }
+			buf2[2 * x - 1] = '\0';
+			if (flag)
+			  {
+			    snprintf(kb, sizeof(kb),  "BannerHex/%d", sockets[i].port);
+			    plug_set_key(desc, kb, ARG_STRING, buf2);
+			  }
+
 			buf[x] = '\0';
-			sprintf(kb, "Banner/%d", sockets[i].port);
+			snprintf(kb, sizeof(kb), "Banner/%d", sockets[i].port);
 			plug_set_key(desc, kb, ARG_STRING, buf);
 #ifdef DISPLAY
 			printf("Banner for port %d: %s\n", sockets[i].port, buf);
 #endif
-			
+			x = DIFFTVu(ti, sockets[i].tictac) / 1000;
+			snprintf(kb, sizeof(kb), "TCPScanner/RwTime1000/%d", sockets[i].port);
+			plug_set_key(desc, kb, ARG_INT, (void*) x);
+			snprintf(kb, sizeof(kb), "TCPScanner/RwTime/%d", sockets[i].port);
+			plug_set_key(desc, kb, ARG_INT, (void*)((x + 500) / 1000));
 		      }
 		    else
 #ifdef DEBUG
@@ -522,10 +613,10 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	    }
 	}
 
-      ti = time(NULL);
+      (void) gettimeofday(&ti, NULL);
       timeout_nb = 0;
       for (i = 0; i < open_sock_nb; i ++)
-	if (sockets[i].fd >= 0 && ti - sockets[i].tictac >= read_timeout)
+	if (sockets[i].fd >= 0 && DIFFTV(ti, sockets[i].tictac) >= read_timeout)
 	  {
 	    switch(sockets[i].state)
 	      {
@@ -534,6 +625,8 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 		printf(">> %d: NO BANNER\n", sockets[i].port);
 #endif
 		timeout_nb ++;
+		snprintf(kb, sizeof(kb), "/tmp/NoBanner/%d", sockets[i].port);
+		plug_set_key(desc, kb, ARG_INT, (void *) 1);
 		break;
 	      case GRAB_SOCKET_OPENING:
 #ifdef DISPLAY
@@ -542,7 +635,9 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 		ports_states[sockets[i].port] = GRAB_PORT_FILTERED;
 		filtered_ports_nb ++;
 		untested_ports_nb --;
+#ifdef BLACKLIST_FILTERED_PORTS
 		remember_filtered_port(desc, sockets[i].port);
+#endif
 		break;
 	      }
 	    my_socket_close(sockets[i].fd); sockets[i].fd = -1;
@@ -572,6 +667,63 @@ banner_grab(const struct in_addr *pia, const char* portrange,
 	else
 	  i ++;
     }
+
+ end:
+#ifdef STATISTICS
+  for (i = 0; i < 3; i ++)
+    if (rtt_nb[i] > 0)
+      {
+	char	*key, rep[160];
+	double	mean, sd = -1.0, emax = -1.0;
+
+	/* Convert from micro-seconds to seconds */
+	rtt_sum[i] /= 1e6; rtt_sum2[i] /= 1e12;
+
+	mean = rtt_sum[i] / rtt_nb[i];
+#if 0
+	snprintf(rep, sizeof(rep), "%6g", mean);
+	plug_set_key(desc, "TCPScanner/MeanRTT", ARG_STRING, rep);
+	x = floor(mean * 1000 + 0.5);
+	plug_set_key(desc, "TCPScanner/MeanRTT1000", ARG_INT, (void*) x);
+	/* rtt_max is integer (uS) */
+	plug_set_key(desc, "TCPScanner/MaxRTT1000", ARG_INT,
+		     (void*)((rtt_max[i] + 500)/1000));
+	snprintf(rep, sizeof(rep), "%6g", (rtt_max[i] + 500000) / 1000000);
+	plug_set_key(desc, "TCPScanner/MaxRTT", ARG_STRING, rep);
+#endif
+	if (rtt_nb[i] > 1)
+	  {
+	    sd = sqrt((rtt_sum2[i] / rtt_nb[i] - mean * mean) * rtt_nb[i] / (rtt_nb[i] - 1));
+	    emax = mean + 3 * sd;
+#if 0
+	    snprintf(rep, sizeof(rep), "%6g", sd);
+	    plug_set_key(desc, "TCPScanner/SDRTT", ARG_STRING, rep);
+	    x = floor(sd * 1000 + 0.5);
+	    plug_set_key(desc, "TCPScanner/SDRTT1000", ARG_INT, (void*)x);
+	    snprintf(rep, sizeof(rep), "%6g", emax);
+	    plug_set_key(desc, "TCPScanner/EstimatedMaxRTT", ARG_STRING, rep);
+	    x = floor(emax * 1000 + 0.5);
+	    plug_set_key(desc, "TCPScanner/EstimatedMaxRTT1000", ARG_INT, (void*)x);
+#endif
+	  }
+	fprintf(stderr, "Mean RTT to %s = %g - [%g, %g] - SD = %g - +3SD = %g [%d %s ports]\n", 
+		inet_ntoa(*pia), mean, 
+		rtt_min[i] / 1e6, rtt_max[i] / 1e6,
+		sd, emax, rtt_nb[i], rtt_type[i]);
+      }
+#endif
+#ifdef DISPLAY
+  fprintf(stderr, "On %s: open_ports_nb=%d / closed_ports_nb = %d / filtered_ports_nb=%d\n", inet_ntoa(*pia), open_ports_nb, closed_ports_nb, filtered_ports_nb);
+#endif
+  plug_set_key(desc, "TCPScanner/OpenPortsNb", ARG_INT, (void*)open_ports_nb);
+  plug_set_key(desc, "TCPScanner/ClosedPortsNb", ARG_INT, (void*)closed_ports_nb);
+  plug_set_key(desc, "TCPScanner/FilteredPortsNb", ARG_INT, (void*)filtered_ports_nb);
+#if 1
+  if (total_ports_nb >= 65535)
+    plug_set_key(desc, "Host/full_scan", ARG_INT, (void*) 1);
+#else
+  plug_set_key(desc, "Host/num_ports_scanned", ARG_INT, (void*)total_ports_nb);
+#endif
   return 0;
 }
 
@@ -585,6 +737,7 @@ PlugExport int plugin_run(struct arglist * desc)
   struct in_addr *p_addr;
   int	timeout = 0, max_cnx, min_cnx, safe_checks = 0, x;
 
+
   p = arg_get_value(preferences, "safe_checks");
   if (p != NULL && strcmp(p, "yes") == 0) safe_checks = 1;
 
@@ -608,7 +761,7 @@ PlugExport int plugin_run(struct arglist * desc)
 
     min_cnx = 8 * max_checks;
     if (safe_checks)
-      max_checks = 16 * max_checks;
+      max_cnx = 16 * max_checks;
     else
       max_cnx = 64 * max_checks;
 
@@ -644,7 +797,11 @@ PlugExport int plugin_run(struct arglist * desc)
 
     if (max_sys_fd <= 0)
       {
-	fp = popen("sysctl kern.maxfiles", "r");
+	if ( find_in_path("sysctl", 0) )
+		fp = popen("sysctl kern.maxfiles", "r");
+ 	else
+		fp = NULL;
+
 	if (fp != NULL)
 	  {
 	    fscanf(fp, "%*s = %d", &max_sys_fd);
@@ -695,6 +852,7 @@ PlugExport int plugin_run(struct arglist * desc)
     return -1;
   comm_send_status(globals, arg_get_value(hostinfos, "NAME"),"portscan", 65535, 65535);
   plug_set_key(desc, "Host/scanned", ARG_INT, (void*)1);
+  plug_set_key(desc, "Host/scanners/nessus_tcp_scanner", ARG_INT, (void*)1);
   return 0;
 }
 
